
/******************************************************************************
* MODULE     : conv_apply.cc
* DESCRIPTION: Function application
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <function.hh>
#include <xtable.hh>

generic
evaluator_rep::apply_lambda (generic g, generic lamb, list args) {
  // cout << "Apply lambda " << g << "; " << lamb << ", " << args << lf;
  list    vars= as_list (lamb[2]);
  generic type= lamb[3];
  generic expr= lamb[4];

  list decls;
  if (N (vars) != N (args))
    return mm_exception (g, "Bad number of arguments");
  while (!nil (vars)) {
    generic var= car (vars);
    if (!is_func (var, MM_TYPE, 2))
      return mm_exception (var, "Variable not typed in lambda expression");
    generic decl= list (var[1], value (car (args), var[2]));
    decls= cons (decl, decls);
    vars= cdr (vars);
    args= cdr (args);
  }
  decls= reverse (decls);

  gentable old_env= env;
  env= xtable (as_gentable (lamb[1]), table ());
  declare (decls);
  generic result= evaluate_as (expr, type);
  env= old_env;

  return result;
}

generic
evaluator_rep::evaluate_conv_apply (generic g) {
  generic fun= evaluate (g[1]);
  if (gentype (fun) == MM_EXCEPTION_TYPE) return fun;
  generic arg= evaluate (g[2]);
  if (gentype (arg) == MM_EXCEPTION_TYPE) return fun;
  if (is_func (fun, MM_LAMBDA, 4))
    return as_value (apply_lambda (g, fun, cdr (as_list (arg)))) -> val;
  if (is_function (fun)) {
    function f= as_function (fun);
    return f (arg);
  }
  if (is_func (fun, CONV_LAMBDA, 2)) {
    env (fun[1])= arg;
    return evaluate (fun[2]);
  }
  return mm_exception (fun, "Not a function");
}
