
/******************************************************************************
* MODULE     : apply.cc
* DESCRIPTION: Function application
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <xtable.hh>
#include <function.hh>
#include <converter.hh>

/******************************************************************************
* Implicit conversions for function application
* ----------------------------------------------------------------------------
* The conversion uses a rudimentary caching mechanism.
* However, this mechanism never frees cached values and
* should therefore be improved in the future.
* ----------------------------------------------------------------------------
* The routine 'apply_prog' returns a program which should be evaluated
* in order to perform the required conversions. The returned program
* necessitates additional auxiliary variables which are stored in
* symbols 0#, 1#, ... In particular, 0# contains the function and
* #1 the argument(s). The auxiliary variables are stored in the current
* environment. We might have to create a local sub-environment
* for each function application in the future. New auxiliary variables
* can be introduced in conv_lambda.cc.
******************************************************************************/

static converter cvt;
static table     cvt_cache;

static generic
apply_prog (generic fun_type, generic arg_type) {
  generic pair= list (fun_type, arg_type);
  if (!cvt_cache->contains (pair))
    cvt_cache (pair)= cvt->apply_prog (fun_type, arg_type);
  return cvt_cache [pair];
}

/******************************************************************************
* Function application
******************************************************************************/

static list
get_values (list l) {
  if (nil (l)) return l;
  return cons (as_value (car (l)) -> val, get_values (cdr (l)));
}

static list
get_types (list l) {
  if (nil (l)) return l;
  return cons (as_value (car (l)) -> type, get_types (cdr (l)));
}

generic
evaluator_rep::apply (generic g, value val, list args) {
  // Applies function value to arguments
  // g contains the original expression
  // val contains the function value
  // args containes the evaluated list of arguments

  generic conv= apply_prog (val->type, cons (MM_CROSS, get_types (args)));
  if (nil (conv)) return mm_exception (g, "Invalid function application");
  env (CONV_FUNCTION)= val->val;
  env (CONV_ARGUMENT)= cons (MM_TUPLE, get_values (args));
  return value (evaluate (conv[0]), conv[1]);
}
