
/******************************************************************************
* MODULE     : declare.cc
* DESCRIPTION: Declaring new global or local variables
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>

/******************************************************************************
* Declaring a new variable
******************************************************************************/

generic
evaluator_rep::declare (symbol var, value val) {
  // globally declare the symbol var as val
  // may also overload var

  if (env->contains (var)) {
    if (!is_value (env[var]))
      return mm_exception (var, "Invalid overloading");
    env (var)= overload (as_value (env[var]), val);
  }
  else env (var)= val;
  return VAL_VOID;
}

void
evaluator_rep::declare (list decls) {
  // locally declare a list of variables
  // each item in "decls" is a pair (var, val),
  // where var is a symbol and val a typed value

  while (!nil (decls)) {
    env->write (car (decls));
    decls= cdr (decls);
  }
}

/******************************************************************************
* Declarations
******************************************************************************/

list
flatten_types (list l) {
  if (nil (l)) return l;
  if (is_func (car (l), MM_TYPE, 2))
    return cons (car (l) [2], flatten_types (cdr (l)));
  return cons (car (l), flatten_types (cdr (l)));
}

generic
evaluator_rep::as_declaration (generic decl) {
  // transforms a declaration like
  //   f(...)...(...):R == ... or
  //   f(...)...(...):R := ...
  // into a pair (var, val),
  // where var is the declared variable and val its typed value.

  generic val = decl[2];
  generic var = decl[1][1];
  generic type= decl[1][2];
  bool alias  = is_func (decl, MM_ASSIGN, 2);
  while (is_list (var)) {
    generic dom= cons (MM_CROSS, flatten_types (cdr (as_list (var))));
    val = list (MM_LAMBDA, cdr (as_list (var)), type, val);
    type= list (MM_FUNCTION, dom, type);
    var = var[0];
  }
  val= evaluate_as (val, type);
  if (!is_value (val)) return val;
  val= as_value (val) -> val;
  if (alias) {
    val = list (val);
    type= list (MM_ALIAS, type);
  }
  return list (var, value (val, type));
}
