
/******************************************************************************
* MODULE     : evaluate.cc
* DESCRIPTION: The main evaluation routines
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <mm_int.hh>
#include <keyword.hh>

/******************************************************************************
* The evaluator
******************************************************************************/

generic
evaluator_rep::evaluate (generic g) {
  // cout << "Evaluate " << g << lf;
  switch (gentype (g)) {
  case MM_STRING_TYPE:
    return value (g, MM_STRING);
  case SYMBOL_TYPE:
    if (env->contains (g))
      return env[g];
    else {
      symbol s= as_symbol (g);
      if (is_int (s))
	return value (mm_int (as_int (s)), MM_INT);
      else if (is_mm_string (s))
	return value (as_mm_string (s), MM_STRING);
      else return mm_exception (g, "Unknown identifier");
    }
  case MM_INT_TYPE:
    return value (g, MM_INT);
  case LIST_TYPE:
    {
      generic res= evaluate (car (g));
      switch (gentype (res)) {
      case VALUE_TYPE:
	{
	  value v= as_value (res);
	  generic args= evaluate (cdr (as_list (g)));
	  if (!is_list (args)) return args;
	  return apply (g, v, as_list (args));
	}
      case KEYWORD_TYPE:
	{
	  keyword kw= as_keyword (res);
	  return (this ->* (kw->fn)) (g);
	}
      case MM_EXCEPTION_TYPE:
	return res;
      }
      fatal_error ("Unexpected situation",
		   "evaluator_rep::evaluate",
		   "evaluator.cc");
    }
  case VALUE_TYPE: // used for equality testing in case expressions
    return g;
  default:
    return mm_exception (g, "Syntax error");
  }
}

generic
evaluator_rep::evaluate (list l) {
  if (nil (l)) return l;
  generic head= evaluate (car (l));
  if (gentype (head) != VALUE_TYPE) {
    if (gentype (head) == MM_EXCEPTION_TYPE) return head;
    return mm_exception (car (l), "Does not produce a value");
  }
  generic tail= evaluate (cdr (l));
  if (gentype (tail) != LIST_TYPE) {
    if (nil (tail)) return list (head);
    return tail;
  }
  return cons (head, as_list (tail));
}

generic
evaluator_rep::filter (generic g, generic type) {
  if (gentype (g) != VALUE_TYPE) {
    if (gentype (g) == MM_EXCEPTION_TYPE) return g;
    return mm_exception (g, "Not a value");
  }
  if (as_value (g) -> type != type)
    return mm_exception (g, "Type mismatch");
  return g;
}

/******************************************************************************
* The user interface
******************************************************************************/

evaluator mmeval;

generic
mmx_evaluate (generic g) {
  return mmeval->evaluate (g);
}
