
/******************************************************************************
* MODULE     : initialize.cc
* DESCRIPTION: Initialize the standard routines
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <function.hh>
#include <mm_int.hh>
#include <mm_ostream.hh>
#include <keyword.hh>

/******************************************************************************
* Some fancy test routines
******************************************************************************/

generic
test_print (generic g) {
  (void) g;
  cout << "Test\n";
  return generic ();
}

static int test_count= 10;

generic
test_bool (generic g) {
  (void) g;
  test_count--;
  return mm_int (test_count >= 0);
}

/******************************************************************************
* Initialization
******************************************************************************/

void
evaluator_rep::declare (symbol var, mmx_keyword fn) {
  env (var)= keyword (var, fn);
}

evaluator_rep::evaluator_rep (): env (table ()) {
  initialize_master ();
  list types   = get_extern_types ();
  list bindings= get_extern_bindings ();
  (void) types;
  while (!nil (bindings)) {
    generic binding= car (bindings);
    declare (as_symbol (binding[0]), value (binding[1], binding[2]));
    bindings= cdr (bindings);
  }

  function test1 ("test_print", test_print);
  function test2 ("test_bool", test_bool);
  declare ("test_print", value (test1, function_type (MM_VOID)));
  declare ("test_bool", value (test2, function_type (MM_BOOL)));

  declare ("true", value (mm_int (true), MM_BOOL));
  declare ("false", value (mm_int (false), MM_BOOL));
  declare ("mmout", value (mm_ostream (symbol ("mmout"), cout), MM_OSTREAM));
  declare ("mmerr", value (mm_ostream (symbol ("mmerr"), cerr), MM_OSTREAM));

  declare (MM_ASSIGN, &evaluator_rep::evaluate_assign);
  declare (MM_BEGIN, &evaluator_rep::evaluate_begin);
  declare (MM_CASE, &evaluator_rep::evaluate_case);
  declare (MM_DEFINE, &evaluator_rep::evaluate_define);
  declare (MM_IF, &evaluator_rep::evaluate_if);
  declare (MM_LAMBDA, &evaluator_rep::evaluate_lambda);
  declare (MM_LOOP, &evaluator_rep::evaluate_loop);

  declare (CONV_APPLY, &evaluator_rep::evaluate_conv_apply);
  declare (CONV_CROSS, &evaluator_rep::evaluate_conv_cross);
  declare (CONV_LAMBDA, &evaluator_rep::evaluate_conv_lambda);
  declare (CONV_UNAND, &evaluator_rep::evaluate_conv_unand);
  declare (CONV_UNALIAS, &evaluator_rep::evaluate_conv_unalias);
  declare (CONV_UNCROSS, &evaluator_rep::evaluate_conv_uncross);
}
