
/******************************************************************************
* MODULE     : evaluator.hh
* DESCRIPTION: the mathemagix evaluator
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __EVALUATOR_HH
#define __EVALUATOR_HH
#include <symbol.hh>
#include <table.hh>
#include <list.hh>
#include <value.hh>
#include <mm_exception.hh>
#include <extern.hh>

class value;
class evaluator_rep;
typedef generic (evaluator_rep::*mmx_keyword) (generic);

class evaluator {
protected:
  evaluator_rep* rep;

public:
  inline evaluator ();
  inline evaluator (evaluator_rep* rep2): rep (rep2) {}
  inline friend evaluator as_evaluator (generic g) {
    if (nil (g)) return evaluator ();
    if (gentype (g) != EVALUATOR_TYPE) type_mismatch (g, "evaluator");
    return evaluator ((evaluator_rep*) g.rep); }
  inline operator generic ();
  inline evaluator_rep* operator -> () { return rep; }

  friend class evaluator_rep;
};

class evaluator_rep: generic_rep {
public:
  gentable env;

public:
  evaluator_rep ();

  int  Type ();
  void Print (ostream& out);
  
  // declarations
  generic as_declaration (generic decl);
  generic declare (symbol var, value val);
  void    declare (list decls);

  // initialization
  void declare (symbol var, mmx_keyword fn);

  // function application
  generic apply_lambda (generic g, generic lamb, list args);
  generic apply (generic g, value val, list args);

  // evaluation
  generic evaluate (generic g);
  generic evaluate (list l);
  generic filter (generic g, generic type);
  inline generic evaluate_as (generic g, generic type);

  // keywords
  generic evaluate_assign (generic);
  generic evaluate_begin (generic);
  generic evaluate_case (generic);
  generic evaluate_define (generic);
  generic evaluate_if (generic);
  generic evaluate_lambda (generic);
  generic evaluate_loop (generic);

  generic evaluate_conv_apply (generic);
  generic evaluate_conv_cross (generic);
  generic evaluate_conv_lambda (generic);
  generic evaluate_conv_unand (generic);
  generic evaluate_conv_unalias (generic);
  generic evaluate_conv_uncross (generic);

  friend class evaluator;
};

inline evaluator::evaluator (): rep (new evaluator_rep ()) {}
inline evaluator::operator generic () { return generic (rep); }
inline generic evaluator_rep::evaluate_as (generic g, generic type) {
  return filter (evaluate (g), type); }

list flatten_types (list l);

generic mmx_evaluate (generic g);
void    pp_E0        (generic g);

#endif // __EVALUATOR_HH
