
/******************************************************************************
* MODULE     : pretty_printer.cc
* DESCRIPTION: Output a mathemagix formula as a LaTeX
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <mm_symbs.hh>
#include <list.hh>

bool texmacs_mode= false;
static list sl;

/******************************************************************************
* Subroutines
******************************************************************************/

void
pp_op (char* s) {
  if (texmacs_mode) cout << s;
  else cout << " " << s << " ";
}

void
pp_op (char* s1, char* s2) {
  if (texmacs_mode) cout << s1;
  else cout << s2;
}

/******************************************************************************
* Pretty printing
******************************************************************************/

void pp_E0 (generic g);
void pp_E1 (generic g);
void pp_Lflux (generic g);
void pp_Rflux (generic g);
void pp_F0 (generic g);
void pp_Seqors (generic g);
void pp_Ors (generic g);
void pp_Xors (generic g);
void pp_F1 (generic g);
void pp_Seqands (generic g);
void pp_Ands (generic g);
void pp_F2 (generic g);
void pp_F3 (generic g);
void pp_F4 (generic g);
void pp_T0 (generic g);
void pp_T1 (generic g);
void pp_T2 (generic g);
void pp_T3 (generic g);
void pp_T4 (generic g);
void pp_T5 (generic g);
void pp_R0 (generic g);
void pp_R1 (generic g);
void pp_R2 (generic g);
void pp_R3 (generic g);
void pp_R4 (generic g);
void pp_R5 (generic g);
void pp_Cl (generic g);
void pp_E0s (generic g);

void
pp_E0 (generic g) {
  if (is_func (g, MM_IF, 3)) {
    pp_op ("\\mathbf{if}\\;", "if ");
    pp_E0 (g[1]);
    pp_op ("\\;\\mathbf{then}\\;", " then ");
    pp_E0 (g[2]);
    pp_op ("\\;\\mathbf{else}\\;", " else ");
    pp_F0 (g[3]);
  }
  else if (is_func (g, MM_BEGIN)) {
    if (texmacs_mode) {
      cout << "\\{\\;";
      pp_Cl (cdr (g));
      cout << "\\}";
    }
    else {
      cout << "{" << lf << indent;
      pp_Cl (cdr (g));
      cout << unindent << "}";
    }
  }
  else if (is_func (g, MM_DEFINE, 2)) {
    pp_F0 (g[1]);
    pp_op ("==");
    pp_E0 (g[2]);
  }
  else if (is_func (g, MM_ASSIGN, 2)) {
    pp_F0 (g[1]);
    pp_op (":=");
    pp_E0 (g[2]);
  }
  else if (is_func (g, MM_PLUSASSIGN, 2)) {
    pp_F0 (g[1]);
    pp_op ("+=");
    pp_E0 (g[2]);
  }
  else if (is_func (g, MM_MINUSASSIGN, 2)) {
    pp_F0 (g[1]);
    pp_op ("-=");
    pp_E0 (g[2]);
  }
  else if (is_func (g, MM_TIMESASSIGN, 2)) {
    pp_F0 (g[1]);
    pp_op ("\\times=", " *= ");
    pp_E0 (g[2]);
  }
  else if (is_func (g, MM_OVERASSIGN, 2)) {
    pp_F0 (g[1]);
    pp_op ("/=");
    pp_E0 (g[2]);
  }
  else pp_E1 (g);
}

void
pp_E1 (generic g) {
  if (is_func (g, MM_LESSLESS, 2)) {
    pp_Lflux (g[1]);
    pp_op ("\\ll", " << ");
    pp_F0 (g[2]);
  }
  else if (is_func (g, MM_GTRGTR, 2)) {
    pp_F0 (g[1]);
    pp_op ("\\gg", " >> ");
    pp_Rflux (g[2]);
  }
  else pp_F0 (g);
}

void
pp_Lflux (generic g) {
  if (is_func (g, MM_LESSLESS, 2)) {
    pp_Lflux (g[1]);
    pp_op ("\\ll", " << ");
    pp_F0 (g[2]);
  }
  else pp_F0 (g);
}

void
pp_Rflux (generic g) {
  if (is_func (g, MM_GTRGTR, 2)) {
    pp_F0 (g[1]);
    pp_op ("\\gg", " >> ");
    pp_Rflux (g[2]);
  }
  else pp_F0 (g);
}

void
pp_F0 (generic g) {
  if (is_func (g, MM_IMPLIES, 2)) {
    pp_F1 (g[1]);
    pp_op ("\\Rightarrow", " => ");
    pp_F1 (g[2]);
  }
  else if (is_func (g, MM_EQUIV, 2)) {
    pp_F1 (g[1]);
    pp_op ("\\Leftrightarrow", " <=> ");
    pp_F1 (g[2]);
  }
  else pp_F1 (g);
}

void
pp_F1 (generic g) {
  if (is_func (g, MM_SEQOR, 2)) {
    pp_F2 (g[1]);
    pp_op ("\\;\\mathbf{or}\\;", " or ");
    pp_Seqors (g[2]);
  }
  else if (is_func (g, MM_OR, 2)) {
    pp_F2 (g[1]);
    pp_op ("\\vee", " || ");
    pp_Ors (g[2]);
  }
  else if (is_func (g, MM_XOR, 2)) {
    pp_F2 (g[1]);
    pp_op ("\\triangledown", " ^^ ");
    pp_Xors (g[2]);
  }
  else pp_F2 (g);
}

void
pp_Seqors (generic g) {
  if (is_func (g, MM_SEQOR, 2)) {
    pp_F2 (g[1]);
    pp_op ("\\;\\mathbf{or}\\;", " or ");
    pp_Seqors (g[2]);
  }
  else pp_F2 (g);
}

void
pp_Ors (generic g) {
  if (is_func (g, MM_OR, 2)) {
    pp_F2 (g[1]);
    pp_op ("\\vee", " || ");
    pp_Ors (g[2]);
  }
  else pp_F2 (g);
}

void
pp_Xors (generic g) {
  if (is_func (g, MM_XOR, 2)) {
    pp_F2 (g[1]);
    pp_op ("\\triangledown", " ^^ ");
    pp_Xors (g[2]);
  }
  else pp_F2 (g);
}

void
pp_F2 (generic g) {
  if (is_func (g, MM_SEQAND, 2)) {
    pp_F3 (g[1]);
    pp_op ("\\;\\mathbf{and}\\;", " and ");
    pp_Seqands (g[2]);
  }
  else if (is_func (g, MM_AND, 2)) {
    pp_F3 (g[1]);
    pp_op ("\\wedge", " && ");
    pp_Ands (g[2]);
  }
  else pp_F3 (g);
}

void
pp_Seqands (generic g) {
  if (is_func (g, MM_SEQAND, 2)) {
    pp_F3 (g[1]);
    pp_op ("\\;\\mathbf{and}\\;", " and ");
    pp_Seqands (g[2]);
  }
  else pp_F3 (g);
}

void
pp_Ands (generic g) {
  if (is_func (g, MM_AND, 2)) {
    pp_F3 (g[1]);
    pp_op ("\\wedge", " && ");
    pp_Ands (g[2]);
  }
  else pp_F3 (g);
}

void
pp_F3 (generic g) {
  if (is_func (g, MM_NOT, 1)) {
    pp_op ("\\neg", "!");
    pp_F4 (g[1]);
  }
  else pp_F4 (g);
}

void
pp_F4 (generic g) {
  if (is_func (g, MM_EQUAL, 2)) {
    pp_T0 (g[1]);
    pp_op ("=");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_UNEQUAL, 2)) {
    pp_T0 (g[1]);
    pp_op ("\\neq", " != ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_LESS, 2)) {
    pp_T0 (g[1]);
    pp_op ("<");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_LESSEQ, 2)) {
    pp_T0 (g[1]);
    pp_op ("\\leqslant", " <= ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_GTR, 2)) {
    pp_T0 (g[1]);
    pp_op (">");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_GTREQ, 2)) {
    pp_T0 (g[1]);
    pp_op ("\\geqslant", " >= ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_NOTLESS, 2)) {
    pp_T0 (g[1]);
    pp_op ("\\nless", " !< ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_NOTLESSEQ, 2)) {
    pp_T0 (g[1]);
    pp_op ("\\nleqslant", " !<= ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_NOTGTR, 2)) {
    pp_T0 (g[1]);
    pp_op ("\\ngtr", " !> ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_NOTGTREQ, 2)) {
    pp_T0 (g[1]);
    pp_op ("\\ngeqslant", " !>= ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_HAS, 2)) {
    pp_T0 (g[1]);
    pp_op ("\\;\\mathbf{has}\\;", " has ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MM_HOLDS, 2)) {
    pp_T0 (g[1]);
    pp_op ("\\;\\mathbf{holds}\\;", " holds ");
    pp_T0 (g[2]);
  }
  else pp_T0 (g);
}

void
pp_T0 (generic g) {
  if (is_func (g, MM_TRANSTYPE, 2)) {
    pp_T0 (g[1]);
    pp_op (":>");
    pp_T1 (g[2]);
  }
  else pp_T1 (g);
}

void
pp_T1 (generic g) {
  if (is_func (g, MM_FUNCTION, 2)) {
    pp_T1 (g[1]);
    pp_op ("\\into", " -> ");
    pp_T2 (g[2]);
  }
  else pp_T2 (g);
}

void
pp_T2 (generic g) {
  if (is_func (g, MM_DOTS, 2)) {
    pp_T3 (g[1]);
    pp_op ("\\ldots", "..");
    pp_T3 (g[2]);
  }
  else pp_T3 (g);
}

void
pp_T3 (generic g) {
  if (is_func (g, MM_PLUS, 2)) {
    pp_T3 (g[1]);
    pp_op ("+");
    pp_T4 (g[2]);
  }
  else if (is_func (g, MM_MINUS, 2)) {
    pp_T3 (g[1]);
    pp_op ("-");
    pp_T4 (g[2]);
  }
  else if (is_func (g, MM_OPLUS, 2)) {
    pp_T3 (g[1]);
    pp_op ("\\oplus", " @+ ");
    pp_T4 (g[2]);
  }
  else if (is_func (g, MM_OMINUS, 2)) {
    pp_T3 (g[1]);
    pp_op ("\\ominus", " @- ");
    pp_T4 (g[2]);
  }
  else pp_T4 (g);
}

void
pp_T4 (generic g) {
  if (is_func (g, MM_TIMES, 2)) {
    pp_T4 (g[1]);
    pp_op ("\\*", " * ");
    pp_T5 (g[2]);
  }
  else if (is_func (g, MM_OVER, 2)) {
    if (texmacs_mode) {
      cout << "\\frac{";
      pp_E0 (g[1]);
      cout << "}{";
      pp_E0 (g[2]);
      cout << "}";
    }
    else {
      pp_T4 (g[1]);
      pp_op ("/", " / ");
      pp_T5 (g[2]);
    }
  }
  else if (is_func (g, MM_OTIMES, 2)) {
    pp_T4 (g[1]);
    pp_op ("\\otimes", " @* ");
    pp_T5 (g[2]);
  }
  else if (is_func (g, MM_OOVER, 2)) {
    pp_T4 (g[1]);
    pp_op ("\\oslash", "@/");
    pp_T5 (g[2]);
  }
  else if (is_func (g, MM_COMPOSE, 2)) {
    pp_T4 (g[1]);
    pp_op ("\\circ", " @ ");
    pp_T5 (g[2]);
  }
  else pp_T5 (g);
}

void
pp_T5 (generic g) {
  if (is_func (g, MM_MINUS, 1)) {
    pp_op ("-", "-");
    pp_R0 (g[1]);
  }
  else if (is_func (g, MM_PREINC, 1)) {
    pp_op ("++", "++");
    pp_R0 (g[1]);
  }
  else if (is_func (g, MM_PREDEC, 1)) {
    pp_op ("--", "--");
    pp_R0 (g[1]);
  }
  else if (is_func (g, MM_NUMBER, 1)) {
    pp_op ("\\#", "#");
    pp_R0 (g[1]);
  }
  else pp_R0 (g);
}

void
pp_R0 (generic g) {
  if (is_func (g, MM_DOLLAR, 1)) {
    pp_op ("\\$", "$");
    pp_R1 (g[1]);
  }
  else pp_R1 (g);
}

void
pp_R1 (generic g) {
  if (is_func (g, MM_POWER, 2)) {
    if (texmacs_mode) {
      pp_R2 (g[1]);
      cout << "^{";
      pp_E0 (g[2]);
      cout << "}";
    }
    else {
      pp_R2 (g[1]);
      pp_op ("^", "^");
      pp_R2 (g[2]);
    }
  }
  else pp_R2 (g);
}

void
pp_R2 (generic g) {
  if (is_func (g, MM_POSTINC, 1)) {
    pp_R2 (g[1]);
    pp_op ("++", "++");
  }
  else if (is_func (g, MM_POSTDEC, 1)) {
    pp_R2 (g[1]);
    pp_op ("--", "--");
  }
  else if (is_func (g, MM_FACTORIAL, 1)) {
    pp_R2 (g[1]);
    pp_op ("!", "!");
  }
  else pp_R3 (g);
}

void
pp_R3 (generic g) {
  if (is_func (g, MM_DOT, 2)) {
    pp_R3 (g[1]);
    pp_op (".", ".");
    pp_R4 (g[2]);
  }
  else if (is_func (g, MM_ACCESS) && (N(g)>1)) {
    if (texmacs_mode) {
      pp_R4 (g[1]);
      cout << "_{";
      pp_E0s (cdr (cdr (g)));
      cout << "}";
    }
    else {
      pp_R4 (g[1]);
      cout << "[";
      pp_E0s (cdr (cdr (g)));
      cout << "]";
    }
  }
  else pp_R4 (g);
}

void
pp_R4 (generic g) {
  if (is_func (g, MM_SQTUPLE)) {
    pp_op ("\\left[", "[");
    pp_E0s (cdr (g));
    pp_op ("\\right]", "]");
  }
  else if (is_func (g, MM_TUPLE)) {
    pp_op ("\\left(", "(");
    pp_E0s (cdr (g));
    pp_op ("\\right)", ")");
  }
  else pp_R5 (g);
}

bool
apply_test (generic g) {
  if (N(g)>0) return true;
  if (gentype (g) == SYMBOL_TYPE) return apply_test (S(g));
  if (gentype (g) == MM_STRING_TYPE) {
    mm_string s= as_mm_string (g);
    if (N(s)==0) return false;
    if ((s[0]>='0') && (s[0]<='9')) return true;
    if ((s[0]=='\042') && (s[N(s)-1]=='\042')) return true;
    int i;
    for (i=0; i<N(s); i++)
      if (((s[i]<'0') || (s[i]>'9')) &&
	  ((s[i]<'a') || (s[i]>'z')) &&
	  ((s[i]<'A') || (s[i]>'Z')) &&
	  ((s[i]<'_') || (s[i]>'_')))
	return false;
    return true;
  }
  return false;
}

void
pp_R5 (generic g) {
  if (N(g) == 0) {
    if ((gentype (g) == SYMBOL_TYPE) &&
	(gentype (S(g)) == MM_STRING_TYPE))
      {
	mm_string s= as_mm_string (S(g));
	if (N(s) == 0) {
	  if (texmacs_mode) cout << "{\\epsilon}";
	  else cout << "epsilon";
	  return;
	}
	if (N(s) == 1) {
	  if (texmacs_mode) cout << "{" << s[0] << "}";
	  else cout << s[0];
	  return;
	}
	if ((s[0]>='0') && (s[0]<='9')) {
	  if (texmacs_mode) cout << "{" << g << "}";
	  else cout << g;
	  return;
	}
      }
    if (texmacs_mode) cout << "\\mathrm{" << g << "}";
    else cout << g;
  }
  else if (is_func (g, MM_PRIME, 1)) {
    pp_R5 (g[1]);
    pp_op ("'", "'");
  }
  else if (apply_test (g[0]) || (!nil (sl)) && (car (sl) == g)) {
    pp_R5  (car (g));
    pp_op ("\\left(", "(");
    pp_E0s (cdr (g));
    pp_op ("\\right)", ")");
  }
  else {
    pp_op ("\\left(", "(");
    sl= cons (g, sl);
    pp_E0 (g);
    sl= cdr (sl);
    pp_op ("\\right)", ")");
  }
}

void
pp_Cl (generic g) {
  if (nil (g)) return;
  pp_E0 (car (g));
  if (texmacs_mode) cout << ";\\;";
  else cout << ";" << lf;
  pp_Cl (cdr (g));
}

void
pp_E0s (generic g) {
  if (N(g) == 1) pp_E0 (car (g));
  if (N(g) <= 1) return;
  pp_E0 (car (g));
  pp_op (", ", ", ");
  pp_E0s (cdr (g));
}
