
/******************************************************************************
* MODULE     : base_types.cc
* DESCRIPTION: Basic types for the Mathemagix interpreter
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extern.hh>
#include <mm_int.hh>
#include <mm_ostream.hh>

/******************************************************************************
* Booleans
******************************************************************************/

generic
bool_not (generic g) {
  return mm_int (!as_mm_int (g[1]));
}

generic
bool_or (generic g) {
  return mm_int (as_mm_int (g[1]) | as_mm_int (g[2]));
}

generic
bool_xor (generic g) {
  return mm_int (as_mm_int (g[1]) ^ as_mm_int (g[2]));
}

generic
bool_and (generic g) {
  return mm_int (as_mm_int (g[1]) & as_mm_int (g[2]));
}

generic
bool_implies (generic g) {
  return mm_int (as_mm_int (g[2]) | (!as_mm_int (g[2])));
}

generic
bool_equivalent (generic g) {
  return mm_int (as_mm_int (g[1]) == as_mm_int (g[2]));
}

generic
bool_equal (generic g) {
  return mm_int (as_mm_int (g[1]) == as_mm_int (g[2]));
}

generic
bool_not_equal (generic g) {
  return mm_int (as_mm_int (g[1]) != as_mm_int (g[2]));
}

generic
bool_output (generic g) {
  mm_ostream mm_out= as_mm_ostream (g[1]);
  ostream& out (mm_out->out);
  if (as_mm_int (g[2])) out << "true"; else out << "false";
  return mm_out;
}

/******************************************************************************
* Integers
******************************************************************************/

generic
int_plus (generic g) {
  return mm_int (as_mm_int (g[1]) + as_mm_int (g[2]));
}

generic
int_minus (generic g) {
  return mm_int (as_mm_int (g[1]) - as_mm_int (g[2]));
}

generic
int_times (generic g) {
  return mm_int (as_mm_int (g[1]) * as_mm_int (g[2]));
}

generic
int_equal (generic g) {
  return mm_int (as_mm_int (g[1]) == as_mm_int (g[2]));
}

generic
int_not_equal (generic g) {
  return mm_int (as_mm_int (g[1]) != as_mm_int (g[2]));
}

generic
int_less (generic g) {
  return mm_int (as_mm_int (g[1]) < as_mm_int (g[2]));
}

generic
int_less_equal (generic g) {
  return mm_int (as_mm_int (g[1]) <= as_mm_int (g[2]));
}

generic
int_greater (generic g) {
  return mm_int (as_mm_int (g[1]) > as_mm_int (g[2]));
}

generic
int_greater_equal (generic g) {
  return mm_int (as_mm_int (g[1]) >= as_mm_int (g[2]));
}

generic
int_output (generic g) {
  mm_ostream mm_out= as_mm_ostream (g[1]);
  ostream& out (mm_out->out);
  out << as_mm_int (g[2]);
  return mm_out;
}

/******************************************************************************
* Strings
******************************************************************************/

generic
string_concat (generic g) {
  return mm_string (as_mm_string (g[1]) * as_mm_string (g[2]));
}

generic
string_equal (generic g) {
  return mm_int (as_mm_string (g[1]) == as_mm_string (g[2]));
}

generic
string_not_equal (generic g) {
  return mm_int (as_mm_string (g[1]) != as_mm_string (g[2]));
}

generic
string_less (generic g) {
  return mm_int (as_mm_string (g[1]) < as_mm_string (g[2]));
}

generic
string_less_equal (generic g) {
  return mm_int (as_mm_string (g[1]) <= as_mm_string (g[2]));
}

generic
string_greater (generic g) {
  return mm_int (as_mm_string (g[2]) < as_mm_string (g[1]));
}

generic
string_greater_equal (generic g) {
  return mm_int (as_mm_string (g[2]) <= as_mm_string (g[1]));
}

generic
string_output (generic g) {
  mm_ostream mm_out= as_mm_ostream (g[1]);
  ostream& out (mm_out->out);
  mm_string s= as_mm_string (g[2]);
  int i, n= N(s);
  for (i=0; i<n; i++) {
    if ((s[i]=='\\') && (i<n-1)) {
      i++;
      if (s[i]=='b') out << '\b';
      else if (s[i]=='n') out << '\n';
      else if (s[i]=='t') out << '\t';
      else out << '\\' << s[i];
    }
    else out << s[i];
  }
  return mm_out;
}

/******************************************************************************
* Interface
******************************************************************************/

void
initialize_base_types () {
  generic MM_BOOL_BOOL= function_type (MM_BOOL, MM_BOOL);
  generic MM_BOOL_BOOL_BOOL= function_type (MM_BOOL, MM_BOOL, MM_BOOL);
  generic MM_INT_INT_INT= function_type (MM_INT, MM_INT, MM_INT);
  generic MM_INT_INT_BOOL= function_type (MM_INT, MM_INT, MM_BOOL);
  generic MM_STRING_STRING_STRING=
    function_type (MM_STRING, MM_STRING, MM_STRING);
  generic MM_STRING_STRING_BOOL= function_type (MM_STRING, MM_STRING, MM_BOOL);

  declare (MM_NOT, bool_not, MM_BOOL_BOOL);
  declare (MM_OR, bool_or, MM_BOOL_BOOL_BOOL);
  declare (MM_XOR, bool_xor, MM_BOOL_BOOL_BOOL);
  declare (MM_AND, bool_and, MM_BOOL_BOOL_BOOL);
  declare (MM_IMPLIES, bool_implies, MM_BOOL_BOOL_BOOL);
  declare (MM_EQUIV, bool_equivalent, MM_BOOL_BOOL_BOOL);
  declare (MM_EQUAL, bool_equal, MM_BOOL_BOOL_BOOL);
  declare (MM_UNEQUAL, bool_not_equal, MM_BOOL_BOOL_BOOL);
  declare (MM_LESSLESS, bool_output,
	   function_type (MM_OSTREAM, MM_BOOL, MM_OSTREAM));

  declare (MM_PLUS, int_plus, MM_INT_INT_INT);
  declare (MM_MINUS, int_minus, MM_INT_INT_INT);
  declare (MM_TIMES, int_times, MM_INT_INT_INT);
  declare (MM_EQUAL, int_equal, MM_INT_INT_BOOL);
  declare (MM_UNEQUAL, int_not_equal, MM_INT_INT_BOOL);
  declare (MM_LESS, int_less, MM_INT_INT_BOOL);
  declare (MM_LESSEQ, int_less_equal, MM_INT_INT_BOOL);
  declare (MM_GTR, int_greater, MM_INT_INT_BOOL);
  declare (MM_GTREQ, int_greater_equal, MM_INT_INT_BOOL);
  declare (MM_LESSLESS, int_output,
	   function_type (MM_OSTREAM, MM_INT, MM_OSTREAM));

  declare (MM_TIMES, string_concat, MM_STRING_STRING_STRING);
  declare (MM_EQUAL, string_equal, MM_STRING_STRING_BOOL);
  declare (MM_UNEQUAL, string_not_equal, MM_STRING_STRING_BOOL);
  declare (MM_LESS, string_less, MM_STRING_STRING_BOOL);
  declare (MM_LESSEQ, string_less_equal, MM_STRING_STRING_BOOL);
  declare (MM_GTR, string_greater, MM_STRING_STRING_BOOL);
  declare (MM_GTREQ, string_greater_equal, MM_STRING_STRING_BOOL);
  declare (MM_LESSLESS, string_output,
	   function_type (MM_OSTREAM, MM_STRING, MM_OSTREAM));
}
