
/******************************************************************************
* MODULE     : error.cc
* DESCRIPTION: Errors and warnings
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <error.hh>
#include <file.hh>

void
print_error_location (ostream& out, generic src) {
  mm_string name1, name2, src1, src2;
  int row1, row2, col1, col2;
  find_location (src->begin, src1, name1, row1, col1);
  find_location (src->end  , src2, name2, row2, col2);
  out << symbol (name1) << ":(" << row1 << "," << col1;
  if (name2 != name1) out << "-...): ";
  else {
    if (row2 == row1) out << "-" << col2 << "): ";
    else out << ")-(" << row2 << "," << col2 << "): ";
  }
}

static void
find_line (int pos, mm_string& line, int& col) {
  mm_string src, name; int row;
  find_location (pos, src, name, row, col);
  if (src == "") line= "";
  else {
    int i= pos- src->begin, i1, i2, n= N(src);
    for (i1= i; (i1>0) && (src[i1-1]!='\n'); i1--);
    for (i2= i; (i2<n) && (src[i2]!='\n'); i2++);
    line= src (i1, i2);
  }
}

void
print_error_source (ostream& out, generic src) {
  mm_string line1, line2;
  int col1, col2;
  find_line (src->begin, line1, col1); col1--;
  find_line (src->end  , line2, col2); col2-=2;

  int i;
  out << indent;
  if ((line1 == line2) &&
      (line1->begin == line2->begin) &&
      (line1->end == line2->end))
    {
      out << symbol (line1) << lf;
      for (i=0; i<col1; i++) out << " ";
      out << "^"; i++;
      for (; i<col2; i++) out << ".";
      if (i==col2) out << "^";
      out << lf;
    }
  else {
    out << symbol (line1) << lf;
    for (i=0; i<col1; i++) out << " ";
    out << "^"; i++;
    for (; i<N(line1); i++) out << ".";
    out << lf;
    for (i=0; i<col2; i++) out << ".";
    if (i==col2) out << "v";
    out << lf;
    out << symbol (line2) << lf;
  }
  out << unindent;
}

static list
merge (list l1, list l2) {
  if (nil (l1)) return l2;
  if (nil (l2)) return l1;
  generic car1= car (l1);
  generic car2= car (l2);
  if (nil (car1)) return cons (car1, merge (cdr (l1), l2));
  if (nil (car2)) return cons (car2, merge (l1, cdr (l2)));
  if ((car1->end < car2->end) ||
      ((car1->end == car2->end) && (car1->begin >= car2->begin)))
    return cons (car1, merge (cdr (l1), l2));
  else return cons (car2, merge (l1, cdr (l2)));
}

static list
sort_errors (list l) {
  int n= N(l);
  if (n<=1) return l;
  else {
    int h= n>>1;
    list l1= sort_errors (l (0, h));
    list l2= sort_errors (l (h, n));
    return merge (l1, l2);
  }
}

void
print_errors (list l) {
  l= sort_errors (l);
  while (!nil (l)) {
    cout << car (l);
    l= cdr (l);
  }
}
