
/******************************************************************************
* MODULE     : proof.hh
* DESCRIPTION: proofs of type conversions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __PROOF_HH
#define __PROOF_HH
#include <generic.hh>
#include <symbol.hh>
#include <list.hh>
#include <table.hh>

class proof_rep: public generic_rep {
public:
  generic f;      // the formula
  generic rule;   // inference rule
  list    deps;   // dependencies on other proofs
  table   inst;   // instantiations for the free variables

protected:
  int Type ();
  int Hash ();
  bool Equal (generic g);
  generic Copy ();
  int Arity ();
  void Print (ostream& out);

public:
  inline proof_rep (generic f2, generic rule2, list deps2):
    f (f2), rule (rule2), deps (deps2) {}
  inline proof_rep (generic f2, generic rule2, list deps2, table inst2):
    f (f2), rule (rule2), deps (deps2), inst (inst2) {}
  friend class proof;
};

class proof {
protected:
  proof_rep* rep;

public:
  inline proof (): rep (NULL) {}
  inline proof (proof_rep* rep2): rep (rep2) {}
  inline proof (generic f, generic rule, list deps, table t):
    rep (new proof_rep (f, rule, deps, t)) {}
  proof (generic from, generic to, table t);
  proof (generic from, generic to, mm_string rule, list deps);
  proof (generic from, generic to, generic rule, list deps);
  inline friend proof as_proof (generic g) {
    if (nil (g) || (gentype (g) != PROOF_TYPE)) type_mismatch (g, "proof");
    return proof ((proof_rep*) g.rep); }
  inline operator generic () { return generic (rep); }
  inline proof_rep* operator -> () { return rep; }

  inline bool operator == (proof p) {
    return
      (rep->f == p->f) &&
      (rep->rule == p->rule) &&
      (rep->deps == p->deps); }
  inline bool operator != (proof p) {
    return
      (rep->f != p->f) ||
      (rep->rule != p->rule) ||
      (rep->deps != p->deps); }
  friend proof copy (proof f);
  friend ostream& operator << (ostream& out, proof f);
  friend class proof_rep;
};

proof ambiguous_proof (generic what, list sols);
inline bool is_ambiguous (proof p) { return p->rule == symbol ("ambiguous"); }
proof substitute (proof p, generic what, generic by);
proof requantify (proof p);
list requantify (list l);

#endif // __PROOF_HH
