
/******************************************************************************
* MODULE     : generic.cc
* DESCRIPTION: Abstract types
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <generic.hh>
#include <symbol.hh>

/******************************************************************************
* Basic routines
******************************************************************************/

int
generic_rep::Type () {
  return UNKNOWN_TYPE;
}

int
generic_rep::Hash () {
  return ((int) this) >> 2;
}

bool
generic_rep::Equal (generic g) {
  return ((void*) this) == ((void*) g.rep);
}

generic
generic_rep::Copy () {
  return generic (this);
}

int
generic_rep::Arity () {
  return 0;
}

generic&
generic_rep::Access (int i) {
  fatal_error ("no data access function associated to generic",
	       "generic_rep::Access", "generic.cc");
  exit (1);
}

generic
generic_rep::Range (int begin, int end) {
  fatal_error ("no range function associated to generic",
	       "generic_rep::Range", "generic.cc");
  exit (1);
}

generic&
generic_rep::Car () {
  fatal_error ("no data access function associated to generic",
	       "generic_rep::Car", "generic.cc");
  exit (1);
}

generic
generic_rep::Cdr () {
  fatal_error ("no range function associated to generic",
	       "generic_rep::Cdr", "generic.cc");
  exit (1);
}

void
generic_rep::Print (ostream& out) {
  out << "#unknown";
}

/******************************************************************************
* Pritty printing
******************************************************************************/

static int cur_indent= 0;
static bool pritty_print = false;
static char indent_s[]  = "| ";
static char unindent_s[]= "\b\b";
static char hrule_s[]= "-----------------------------------------------------------------------------\n";

ostream&
operator << (ostream& out, print_format x) {
  int i;
  switch (x) {
  case blank:
    indent_s[0]= ' ';
    break;
  case stroke:
    indent_s[0]= '|';
    break;
  case indent:
    cur_indent++;
    cout << indent_s;
    break;
  case unindent:
    cur_indent--;
    cout << unindent_s;
    break;
  case lf:
    out << "\n";
    for (i=0; i<cur_indent; i++)
      out << indent_s;
    break;
  case pritty:
    pritty_print = true;
    break;
  case normal:
    pritty_print = false;
    break;
  case hrule:
    out << hrule_s;
    break;
  }
  return out;
}

void
print_entries (ostream& out, generic l, bool flag) {
  int n= N(l);
  if ((n == 0) && (!nil (l))) out << "{" << l << "}";
  else {
    out << (flag? "(": "<");
    if (pritty_print) {
      if (n != 0) {
	if (N(car(l)) == 0) {
	  out << car (l);
	  l= cdr (l); n--;
	}
	cur_indent++;
	while (n != 0) {
	  out << lf << car (l);
	  l= cdr (l); n--;
	}
	cur_indent--;
      }
    }
    else while (n != 0) {
      out << car (l);
      l= cdr (l); n--;
      if (n != 0) out << " ";
    }
    out << (flag? ")": ">");
  }
}

/******************************************************************************
* Other routines
******************************************************************************/

generic&
S (generic g) {
  if (gentype (g) == SYMBOL_TYPE) return S (as_symbol (g));
  type_mismatch (g, "symbol", "S", "generic.cc");
  exit (1);
}

generic
extend (generic g, generic with) {
  if (nil (g) || nil (with)) return g;
  if ((with->begin == -1) || (with->end == -1)) return g;
  if ((g->begin == -1) || (g->end == -1)) {
    g->begin= with->begin;
    g->end  = with->end  ;
  }
  else {
    g->begin= (g->begin < with->begin ? g->begin : with->begin);
    g->end  = (g->end   > with->end   ? g->end   : with->end  );
  }
  return g;
}

void
fatal_error (char* message, char* routine, char* file) {
  cout << "\nFatal error: " << message
       << "\n             in '" << routine
       << "' in file '" << file << "'\n";
  exit (1);
}

void
type_mismatch (generic g, char* expected_type) {
  cout << "\nFatal error: expected a value of type " << expected_type
       << "\n             in 'as_" << expected_type
       << "' in file '" << expected_type << ".hh'"
       << "\nValue was  : " << g
       << "\nType was   : " << gentype (g) << "\n";
  exit (1);
}

void
type_mismatch (generic g, char* expected_type, char* routine, char* file) {
  cout << "\nFatal error: expected a value of type " << expected_type
       << "\n             in '" << routine
       << "' in file '" << file << "'"
       << "\nValue was  : " << g
       << "\nType was   : " << gentype (g) << "\n";
  exit (1);  
}
