
/******************************************************************************
* MODULE     : mm_string.hh
* DESCRIPTION: Fixed size mm_strings. Zero-characters are allowed in mm_strings.
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __STRING_HH
#define __STRING_HH
#include <generic.hh>

class mm_string_rep: public generic_rep {
private:
  int n;
  char* a;

  int Type ();
  int Hash ();
  bool Equal (generic g);
  generic Copy ();
  void Print (ostream& out);

public:
  inline mm_string_rep (): n(0), a(NULL) {}
         mm_string_rep (int n);
  inline ~mm_string_rep () { if (n!=0) delete[] a; }
  void resize (int n2);

  friend class mm_string;
  friend inline int N (mm_string s);
  friend inline char* S (mm_string s);
};

class mm_string {
private:
  mm_string_rep* rep;
  inline mm_string (mm_string_rep* rep2): rep (rep2) {}

public:
  inline mm_string (): rep (new mm_string_rep()) {}
  inline mm_string (int n): rep (new mm_string_rep (n)) {}
  mm_string (char c);
  mm_string (char *s);
  mm_string (char *s, int n);
  inline friend mm_string as_mm_string (generic g) {
    if (gentype(g) != MM_STRING_TYPE) type_mismatch (g, "mm_string");
    return mm_string ((mm_string_rep*) g.rep); }
  inline operator generic () { return generic (rep); }
  inline mm_string_rep* operator -> () { return rep; }

  inline friend int N (mm_string s) { return s.rep->n; }
  inline friend char* S (mm_string s) { return s.rep->a; }
  inline char& operator [] (int i) { return rep->a[i]; }
  mm_string operator () (int start, int end);
  bool operator == (char* s);
  bool operator != (char* s);
  bool operator == (mm_string s);
  bool operator != (mm_string s);
  friend mm_string copy (mm_string s);
  friend mm_string as_mm_string (int i);
  friend char*  as_charp (mm_string s);
  friend ostream& operator << (ostream& out, mm_string s);

  friend class mm_string_rep;
};

mm_string&  operator << (mm_string& s, char c);
mm_string&  operator << (mm_string& s, mm_string t);
mm_string&  operator << (mm_string& s, int i);
mm_string   operator * (mm_string s, mm_string t);
bool     operator <  (mm_string s, mm_string t);
bool     operator <= (mm_string s, mm_string t);

#endif // __STRING_HH
